defmodule Pleroma.Uploaders.MFC do
  @behaviour Pleroma.Uploaders.Uploader

  alias __MODULE__.Image
  alias __MODULE__.Video
  alias Pleroma.Upload

  def get_file(file), do: store().get_file(file)

  def put_file(upload, :video_or_gif) do
    original = rename_original_path(upload)

    with {:ok, {:file, path}} <- store().put_file(original, {:acl, :private}),
         {:ok, _} <- Registry.register(__MODULE__, path, path),
         :ok <- Video.convert(Video.Client.client(), path),
         {:ok, conversion_result} <- wait_for_conversion() do
      meta =
        conversion_result
        |> Map.get(:meta, %{})
        |> Map.put("source_key", original.path)

      upload_result = %{
        url_spec: {:file, Map.get(conversion_result, :path)},
        meta: meta
      }

      {:ok, {:upload_result, upload_result}}
    else
      :duplicate ->
        {:ok, {:file, Path.rootname(upload.path) <> ".mp4"}}

      error = {:error, _} ->
        error
    end
  end

  # put video file
  #
  def put_file(%Upload{content_type: "video" <> _} = upload) do
    put_file(upload, :video_or_gif)
  end

  def put_file(%Upload{content_type: "image/gif"} = upload) do
    with {:ok, {:upload_result, upload_result}} <- put_file(upload, :video_or_gif) do
      {:ok,
       {:upload_result,
        %{
          url_spec: upload_result[:url_spec],
          meta: Map.put(upload_result[:meta], "content_type", "image/gifv")
        }}}
    else
      e -> e
    end
  end

  # put image file
  #
  def put_file(%Upload{content_type: "image/" <> _type} = upload) do
    original = rename_original_path(upload)

    with {:ok, {:file, _path}} <- store().put_file(original, {:acl, :private}),
         {:ok, result} <- Image.convert(Image.Client.client(), upload.path) do
      meta =
        result
        |> Map.get("versions", %{})
        |> Map.put("source_key", original.path)

      upload_result = %{
        url_spec: {:file, upload.path},
        meta: meta
      }

      {:ok, {:upload_result, upload_result}}
    else
      :duplicate ->
        {:ok, {:file, upload.path}}

      error = {:error, _} ->
        error
    end
  end

  def put_file(upload) do
    store().put_file(upload)
  end

  def preview_url("video" <> _, url), do: Video.build_preview_url(url)
  def preview_url("image/gif" <> _, url), do: Video.build_preview_url(url)

  def preview_url("image" <> _type, url) do
    Image.build_preview_url(url)
  end

  def preview_url(_type, href), do: href

  def rename_original_path(%Upload{} = upload) do
    %Upload{upload | path: rename_original_path(upload.path)}
  end

  def rename_original_path(path) when is_binary(path) do
    path <> "_original"
  end

  defp store, do: Pleroma.Config.get([__MODULE__, :store], Pleroma.Uploaders.S3)

  def wait_for_conversion do
    receive do
      {__MODULE__, {:ok, path}} ->
        :ok = Registry.unregister(__MODULE__, path)
        {:ok, path}

      {__MODULE__, {:error, error}} ->
        {:error, error}
    after
      conversion_wait() -> {:error, "conversion timeout"}
    end
  end

  defp conversion_wait do
    [__MODULE__, :video_conversion, :conversion_wait]
    |> Pleroma.Config.get(:timer.minutes(2))
  end
end
