defmodule Pleroma.Web.Auth.MfcAuthenticator do
  alias Pleroma.User

  @behaviour Pleroma.Web.Auth.Authenticator
  @behaviour Pleroma.Web.Auth.MongooseIMAuthenticator

  @base Pleroma.Web.Auth.PleromaAuthenticator
  @controller Pleroma.Web.OAuth.OAuthController

  defdelegate get_registration(conn), to: @base

  defdelegate create_from_registration(conn, registration), to: @base

  def get_user(%Plug.Conn{} = conn) do
    {name, password} =
      case conn.params do
        %{"authorization" => %{"name" => name, "password" => password}} ->
          {name, password}

        %{"grant_type" => "password", "username" => name, "password" => password} ->
          {name, password}
      end

    with {_, {:ok, user_data}} <-
           {:mfc_auth,
            Pleroma.Web.Mfc.Login.authenticate(
              name,
              password,
              to_string(:inet.ntoa(conn.remote_ip))
            )},
         {_, true} <-
           {:access_level_check,
            user_data["access_level"] >=
              Application.get_env(:pleroma, :mfc) |> Keyword.get(:minimum_access_level)},
         {_, %User{} = user} <-
           {:user_get,
            Pleroma.Web.Mfc.Utils.get_or_create_mfc_user(
              user_data["user_id"],
              user_data["username"],
              user_data["avatar_url"]
            )},
         {_, user} <-
           {:user_tag,
            User.tag(user, Pleroma.Web.Mfc.Utils.tags_for_level(user_data["access_level"]))} do
      {:ok, user}
    else
      error -> {:error, error}
    end
  end

  def handle_error(%Plug.Conn{params: %{"authorization" => auth_attrs}} = conn, error) do
    case error do
      {:get_user, {:error, {:access_level_check, _}}} ->
        conn
        |> Phoenix.Controller.put_flash(:error, "Only available for Premium accounts")
        |> @controller.authorize(auth_attrs)

      {:get_user, {:error, {:user_get, _}}} ->
        conn
        |> @controller.authorize(Map.put(auth_attrs, "registration", true))

      error ->
        @base.handle_error(conn, error)
    end
  end

  def auth_template, do: "show_mfc.html"

  def oauth_consumer_template, do: nil

  @spec user_exists?(String.t()) :: boolean()
  def user_exists?(username) do
    if Pleroma.User.Query.build(%{nickname: username, local: true, mfc_user: true})
       |> Pleroma.Repo.one(),
       do: true,
       else: false
  end

  @spec check_password(map()) :: boolean() | nil
  def check_password(%{username: username, password: password, ip: ip}) do
    case Pleroma.Web.Mfc.Login.authenticate(
           username,
           password,
           ip
         ) do
      {:ok, _user_data} -> true
      _ -> false
    end
  end
end
