# Pleroma: A lightweight social networking server
# Copyright © 2017-2019 Pleroma Authors <https://pleroma.social/>
# SPDX-License-Identifier: AGPL-3.0-only

defmodule Pleroma.Web.Metadata.Providers.OpenGraph do
  alias Pleroma.User
  alias Pleroma.Web.Metadata
  alias Pleroma.Web.Metadata.Providers.Provider
  alias Pleroma.Web.Metadata.Utils

  @behaviour Provider
  @media_types ["image", "audio", "video"]

  @impl Provider
  def build_tags(%{
        object: object,
        url: url,
        user: user
      }) do
    attachments = build_attachments(object)
    scrubbed_content = Utils.scrub_html_and_truncate(object)

    # Zero width space
    content =
      if scrubbed_content != "" and scrubbed_content != "\u200B" do
        ": “" <> scrubbed_content <> "”"
      else
        ""
      end

    # Most previews only show og:title which is inconvenient. Instagram
    # hacks this by putting the description in the title and making the
    # description longer prefixed by how many likes and shares the post
    # has. Here we use the descriptive nickname in the title, and expand
    # the full account & nickname in the description. We also use the cute^Wevil
    # smart quotes around the status text like Instagram, too.

    title = "#{user.name}" <> content
    description = "#{Utils.user_name_string(user)}" <> content

    data = [
      {:meta, [property: "og:title", content: title], []},
      {:meta, [property: "og:url", content: url], []},
      {:meta, [property: "og:description", content: description], []},
      {:meta, [property: "og:type", content: "website"], []}
    ]

    if attachments == [] or Metadata.activity_nsfw?(object) do
      attachment_url = user |> User.avatar_url() |> Utils.attachment_url()

      data ++
        [
          {:meta, [property: "og:image", content: attachment_url], []},
          {:meta, [property: "og:image:width", content: 150], []},
          {:meta, [property: "og:image:height", content: 150], []}
        ]
    else
      data ++ attachments
    end
  end

  @impl Provider
  def build_tags(%{user: user}) do
    truncated_bio = Utils.scrub_html_and_truncate(user.bio || "")
    attachment_url = user |> User.avatar_url() |> Utils.attachment_url()
    title = Utils.user_name_string(user)

    [
      {:meta, [property: "og:title", content: title], []},
      {:meta, [property: "og:url", content: User.profile_url(user)], []},
      {:meta, [property: "og:description", content: truncated_bio], []},
      {:meta, [property: "og:type", content: "website"], []},
      {:meta, [property: "og:image", content: attachment_url], []},
      {:meta, [property: "og:image:width", content: 150], []},
      {:meta, [property: "og:image:height", content: 150], []}
    ]
  end

  defp build_attachments(%{data: %{"attachment" => attachments}}) do
    Enum.flat_map(attachments, &build_attachment/1)
  end

  defp build_attachments(_), do: []

  defp build_attachment(attachment) do
    Enum.flat_map(attachment["url"], fn url ->
      media_type = Utils.fetch_media_type(@media_types, url["mediaType"])
      url = Utils.attachment_url(url["href"])

      # TODO: Add additional properties to objects when we have the data available.
      # Also, Whatsapp only wants JPEG or PNGs. It seems that if we add a second og:image
      # object when a Video or GIF is attached it will display that in Whatsapp Rich Preview.

      case media_type do
        type when type in ["audio", "video"] ->
          name = "og:" <> media_type
          [{:meta, [property: name, content: url], []}]

        "image" ->
          name = "og:" <> media_type

          [
            {:meta, [property: name, content: url], []},
            {:meta, [property: "og:image:width", content: width(attachment)], []},
            {:meta, [property: "og:image:height", content: height(attachment)], []}
          ]

        _ ->
          []
      end
    end)
  end

  defp width(%{"meta" => %{"original" => %{"width" => width}}}), do: width
  defp width(_), do: 150

  defp height(%{"meta" => %{"original" => %{"height" => height}}}), do: height
  defp height(_), do: 150
end
