defmodule Mix.Tasks.Mfc.ReencodeTest do
  use Pleroma.Web.ConnCase
  import Pleroma.Factory
  import Tesla.Mock
  alias Pleroma.Object
  alias Pleroma.Repo

  describe "reencode" do
    setup do
      user = insert(:user)

      image_meta = %{
        "small" => %{
          "size" => "333x400",
          "width" => 333,
          "aspect" => 0.8325,
          "height" => 400,
          "method" => "resize",
          "dest_key" =>
            "0619b7f00f1e46653bbb6b79f786294696bcdccc7b30f7f4028bbeca7c2e4540.jpg.preview.jpg",
          "content_type" => "image/jpeg"
        },
        "original" => %{
          "size" => "720x866",
          "width" => 720,
          "aspect" => 0.8314087759815243,
          "height" => 866,
          "method" => "resize",
          "dest_key" => "0619b7f00f1e46653bbb6b79f786294696bcdccc7b30f7f4028bbeca7c2e4540.jpg",
          "content_type" => "image/jpeg"
        },
        "source_key" =>
          "0619b7f00f1e46653bbb6b79f786294696bcdccc7b30f7f4028bbeca7c2e4540.jpg_original"
      }

      object_data = %{
        "url" => [
          %{
            "href" =>
              "https://social.dome.com/media/0619b7f00f1e46653bbb6b79f786294696bcdccc7b30f7f4028bbeca7c2e4540.jpg",
            "type" => "Link",
            "mediaType" => "image/jpeg"
          }
        ],
        "meta" => image_meta,
        "name" => "ZDv317OsNojizA.jpg",
        "type" => "Document",
        "actor" => user.ap_id
      }

      {:ok, image_object} = Repo.insert(%Object{data: object_data})

      video_meta = %{
        "fps" => 25,
        "size" => "720x408",
        "small" => %{
          "size" => "720x408",
          "width" => 720,
          "aspect" => 1.7647058823529411,
          "height" => 408
        },
        "width" => 720,
        "aspect" => 1.7647058823529411,
        "height" => 408,
        "duration" => 33.53,
        "original" => %{
          "width" => 720,
          "height" => 408,
          "bitrate" => 543_297,
          "duration" => 33.53,
          "frame_rate" => "25/1"
        },
        "audio_encode" => "aac (mp4a / 0x6134706d)",
        "audio_bitrate" => "44100 Hz",
        "audio_channels" => "stereo",
        "source_key" =>
          "c67a114ceff7817f03dae55fc49c63e691ea38dfd36ba9f761877186cba47f12.mp4_original"
      }

      object_data = %{
        "url" => [
          %{
            "href" =>
              "https://social.dome.com/media/c67a114ceff7817f03dae55fc49c63e691ea38dfd36ba9f761877186cba47f12.mp4",
            "type" => "Link",
            "mediaType" => "video/mp4"
          }
        ],
        "meta" => video_meta,
        "name" => "BMbW85Vze3DYg.mp4",
        "type" => "Document",
        "actor" => user.ap_id
      }

      {:ok, video_object} = Repo.insert(%Object{data: object_data})

      gif_data = %{
        "url" => [
          %{
            "href" =>
              "https://social.dome.com/media/d7c016123f31b86936a2046c42cc8380036fbcd6d008ba0c2801eb73884ce8b9.mp4",
            "type" => "Link",
            "mediaType" => "image/gif"
          }
        ],
        "meta" => %{
          "source_key" =>
            "d7c016123f31b86936a2046c42cc8380036fbcd6d008ba0c2801eb73884ce8b9.mp4_original"
        },
        "name" => "QgBJ-6DXX5xRMw.gif",
        "type" => "Document",
        "actor" => user.ap_id
      }

      {:ok, gif_object} = Repo.insert(%Object{data: gif_data})

      {:ok, image_object: image_object, video_object: video_object, gif_object: gif_object}
    end

    @image_versions %{
      original: %{
        width: 2048,
        height: 2048,
        size: "2048x2048",
        aspect: 1.2484394506866416,
        method: "resize",
        dest_key: "original.png",
        content_type: "image/png"
      },
      small: %{
        width: 800,
        height: 800,
        size: "800x800",
        aspect: 1.2483426823049464,
        method: "smartcrop",
        dest_key: "small.png",
        content_type: "image/png"
      },
      source_key: "0619b7f00f1e46653bbb6b79f786294696bcdccc7b30f7f4028bbeca7c2e4540.jpg_original"
    }

    test "only image", %{image_object: image_object} do
      mock(fn
        %{method: :post, url: "http://test.test/api/v2/images"} ->
          json(%{versions: @image_versions})
      end)

      Mix.Tasks.Mfc.Reencode.run(["images"])

      reloaded = Repo.get(Object, image_object.id)
      meta = reloaded.data["meta"]
      assert meta == to_string_keys(@image_versions)
    end

    @video_meta %{
      "dest_key" => "safe_video_n1.mp4",
      "still_key" => "safe_video.mp4.preview.jpg",
      "meta" => %{
        "duration" => 15.41,
        "bitrate" => 891_066,
        "filesize" => 1_683_002,
        "video_stream" =>
          "h264 (High) (avc1 / 0x31637661), yuv420p, 720x404 [SAR 1:1 DAR 180:101]",
        "video_codec" => "h264",
        "colorspace" => "yuv420p",
        "resolution" => "720x404",
        "width" => 720,
        "height" => 404,
        "frame_rate" => "13500000/451451",
        "audio_stream" => "aac (mp4a / 0x6134706d), 44100 Hz, stereo, fltp, 134472 bit/s",
        "audio_codec" => "aac",
        "audio_sample_rate" => 44_100,
        "audio_channels" => 2,
        "action" => "success"
      },
      "status" => "ok"
    }

    test "only video", %{conn: conn, video_object: video_object} do
      mock(fn
        %{
          method: :post,
          url: "http://test.test/api/v1/videos",
          body:
            "{\"client\":\"xxxx\",\"dest_key\":\"c67a114ceff7817f03dae55fc49c63e691ea38dfd36ba9f761877186cba47f12.mp4\",\"secret\":\"dddd\",\"source_key\":\"c67a114ceff7817f03dae55fc49c63e691ea38dfd36ba9f761877186cba47f12.mp4_original\",\"still_key\":\"c67a114ceff7817f03dae55fc49c63e691ea38dfd36ba9f761877186cba47f12.mp4.preview.jpg\",\"still_seek_percentage\":50}"
        } = env ->
          response = Jason.decode!(env.body)
          video_meta = Map.put(@video_meta, "source_key", response["source_key"])
          post(conn, "/api/pleroma/uploaders/mfc/success", video_meta)
          %Tesla.Env{status: 200, body: %{}}
      end)

      Mix.Tasks.Mfc.Reencode.run(["videos"])

      reloaded_object = Repo.get(Object, video_object.id)

      meta = reloaded_object.data["meta"]

      video_expected = %{
        "duration" => 15.41,
        "height" => 404,
        "width" => 720,
        "audio_channels" => "stereo",
        "aspect" => 1.7821782178217822,
        "audio_bitrate" => "44100 Hz",
        "audio_encode" => "aac (mp4a / 0x6134706d)",
        "fps" => 30,
        "original" => %{
          "bitrate" => 891_066,
          "duration" => 15.41,
          "frame_rate" => "13500000/451451",
          "height" => 404,
          "width" => 720
        },
        "size" => "720x404",
        "small" => %{
          "aspect" => 1.7821782178217822,
          "height" => 404,
          "size" => "720x404",
          "width" => 720
        }
      }

      assert meta ==
               Map.put(
                 video_expected,
                 "source_key",
                 "c67a114ceff7817f03dae55fc49c63e691ea38dfd36ba9f761877186cba47f12.mp4_original"
               )
    end

    @gif_meta %{
      "source_key" =>
        "d7c016123f31b86936a2046c42cc8380036fbcd6d008ba0c2801eb73884ce8b9.mp4_original",
      "dest_key" => "safe_image.mp4",
      "still_key" => "safe_image.gif.preview.jpg",
      "meta" => %{
        "duration" => 15.39,
        "bitrate" => 891_066,
        "filesize" => 1_683_002,
        "video_stream" =>
          "h264 (High) (avc1 / 0x31637661), yuv420p, 720x404 [SAR 1:1 DAR 180:101]",
        "video_codec" => "h264",
        "colorspace" => "yuv420p",
        "resolution" => "720x404",
        "width" => 720,
        "height" => 404,
        "frame_rate" => "13500100/451451"
      },
      "status" => "ok"
    }

    test "only gif", %{conn: conn, gif_object: gif_object} do
      mock(fn
        %{
          method: :post,
          url: "http://test.test/api/v1/videos",
          body:
            "{\"client\":\"xxxx\",\"dest_key\":\"d7c016123f31b86936a2046c42cc8380036fbcd6d008ba0c2801eb73884ce8b9.mp4\",\"secret\":\"dddd\",\"source_key\":\"d7c016123f31b86936a2046c42cc8380036fbcd6d008ba0c2801eb73884ce8b9.mp4_original\",\"still_key\":\"d7c016123f31b86936a2046c42cc8380036fbcd6d008ba0c2801eb73884ce8b9.mp4.preview.jpg\",\"still_seek_percentage\":50}"
        } = env ->
          response = Jason.decode!(env.body)
          gif_meta = Map.put(@gif_meta, "source_key", response["source_key"])
          post(conn, "/api/pleroma/uploaders/mfc/success", gif_meta)
          %Tesla.Env{status: 200, body: %{}}
      end)

      Mix.Tasks.Mfc.Reencode.run(["gifs"])

      reloaded_object = Repo.get(Object, gif_object.id)

      meta = reloaded_object.data["meta"]

      gif_expected = %{
        "duration" => 15.39,
        "height" => 404,
        "width" => 720,
        "aspect" => 1.7821782178217822,
        "fps" => 30,
        "original" => %{
          "bitrate" => 891_066,
          "duration" => 15.39,
          "frame_rate" => "13500100/451451",
          "height" => 404,
          "width" => 720
        },
        "size" => "720x404",
        "small" => %{
          "aspect" => 1.7821782178217822,
          "height" => 404,
          "size" => "720x404",
          "width" => 720
        },
        "source_key" =>
          "d7c016123f31b86936a2046c42cc8380036fbcd6d008ba0c2801eb73884ce8b9.mp4_original"
      }

      assert meta == gif_expected
    end

    test "limit option", %{image_object: image_object} do
      user = insert(:user)

      image_meta = %{
        "small" => %{
          "size" => "333x400",
          "width" => 333,
          "aspect" => 0.8325,
          "height" => 400,
          "method" => "resize",
          "dest_key" => "12312312312.jpg.preview.jpg",
          "content_type" => "image/jpeg"
        },
        "original" => %{
          "size" => "720x866",
          "width" => 720,
          "aspect" => 0.8314087759815243,
          "height" => 866,
          "method" => "resize",
          "dest_key" => "12312312312.jpg",
          "content_type" => "image/jpeg"
        },
        "source_key" => "another.jpg_original"
      }

      object_data = %{
        "id" => Pleroma.Web.ActivityPub.Utils.generate_object_id(),
        "url" => [
          %{
            "href" => "https://social.dome.com/media/12312312312.jpg",
            "type" => "Link",
            "mediaType" => "image/jpeg"
          }
        ],
        "meta" => image_meta,
        "name" => "ZDv317OsNojizA.jpg",
        "type" => "Document",
        "actor" => user.ap_id
      }

      {:ok, object2} = Repo.insert(%Object{data: object_data})

      assert length(Repo.all(Object)) == 4

      mock(fn
        %{method: :post, url: "http://test.test/api/v2/images"} ->
          json(%{versions: @image_versions})
      end)

      Mix.Tasks.Mfc.Reencode.run(["images", "--limit", "1"])

      reloaded = Repo.get(Object, image_object.id)
      reloaded2 = Repo.get(Object, object2.id)

      assert reloaded.data["meta"] == to_string_keys(@image_versions)

      assert image_meta == reloaded2.data["meta"]
    end
  end

  defp to_string_keys(data) when is_map(data) do
    for {k, v} <- data, into: %{}, do: {to_string(k), to_string_keys(v)}
  end

  defp to_string_keys(data) when is_integer(data) or is_binary(data) or is_float(data), do: data
end
