defmodule Pleroma.Web.Mfc.ApiTest do
  use Pleroma.DataCase
  import Pleroma.Factory
  alias Pleroma.Config
  alias Pleroma.Mfc.Notify
  alias Pleroma.Object
  alias Pleroma.Web.CommonAPI
  alias Pleroma.Web.Mfc.Api

  test "does a signed request" do
    data = %{x: 1, y: 2}
    url = Pleroma.Config.get([:mfc, :account_creation_endpoint])

    Tesla.Mock.mock(fn %{url: ^url, body: body} ->
      assert body =~ "x=1"
      assert body =~ "y=2"
      assert body =~ "gh="
      assert body =~ "service_id="
      assert body =~ "t="

      send(self(), :called_api)
      %Tesla.Env{status: 200}
    end)

    Api.authenticated_request(:post, url, data)

    assert_received(:called_api)
  end

  test "it fetches the users another user is following" do
    body =
      "{\"err\":0,\"data\":[{\"id\":372,\"nick\":\"Tenshi\",\"access_level\":1},{\"id\":18062,\"nick\":\"CornDogLover\",\"access_level\":1}]}"

    url = Config.get([:mfc, :following_endpoint_v2])

    Tesla.Mock.mock(fn %{url: ^url} ->
      %Tesla.Env{status: 200, body: body}
    end)

    user_ids = Api.get_following_for_mfc_id("123")

    assert "18062" in user_ids
    assert "372" in user_ids
  end

  test "it calls the account creation endpoint" do
    url = Config.get([:mfc, :account_creation_endpoint])
    user = insert(:user, %{mfc_id: "1"})
    expected_time = user.inserted_at |> DateTime.from_naive!("Etc/UTC") |> DateTime.to_unix()

    Tesla.Mock.mock(fn %{url: ^url, body: body} ->
      assert body =~ "mfc_id=1"

      assert body =~ "username=#{user.nickname}"

      assert body =~ "social_created_at=#{expected_time}"

      send(self(), :called_api)
      %Tesla.Env{status: 200}
    end)

    assert Notify.perform(:account_creation, user)
    assert_received(:called_api)
  end

  test "it calls the status deletion endpoint" do
    url = Config.get([:mfc, :status_deletion_endpoint])
    user = insert(:user, %{mfc_id: "1"})
    mod = insert(:user, %{mfc_id: "2"})

    {:ok, activity} = CommonAPI.post(user, %{"status" => "Hello"})
    expected_time = activity.inserted_at |> DateTime.from_naive!("Etc/UTC") |> DateTime.to_unix()

    Tesla.Mock.mock(fn %{url: ^url, body: body} ->
      send(self(), :called_api)
      assert body =~ "mfc_id=1"
      assert body =~ "deleted_by_id=2"
      assert body =~ "post_id=#{activity.id}"
      assert body =~ "deleted_at=#{expected_time}"

      %Tesla.Env{status: 200}
    end)

    assert Notify.perform(:status_delete, activity, mod, activity.inserted_at)
    assert_received(:called_api)
  end

  test "it calls the status creation endpoint" do
    url = Config.get([:mfc, :status_creation_endpoint])
    user = insert(:user, %{mfc_id: "1"})
    {:ok, activity} = CommonAPI.post(user, %{"status" => "Hello"})
    object = Object.normalize(activity)

    expected_time = activity.inserted_at |> DateTime.from_naive!("Etc/UTC") |> DateTime.to_unix()

    Tesla.Mock.mock(fn %{url: ^url, body: body} ->
      assert body =~ "mfc_id=1"
      assert body =~ "last_post_date=#{expected_time}"
      assert body =~ "last_post_content=#{object.data["content"]}"
      assert body =~ "status_count=1"
      assert body =~ "last_post_id=#{activity.id}"
      assert body =~ "last_post_url=#{URI.encode_www_form(object.data["id"])}"
      send(self(), :called_api)

      %Tesla.Env{status: 200}
    end)

    assert Notify.perform(:status_create, activity)
    assert_received(:called_api)

    {:ok, reply} =
      CommonAPI.post(user, %{
        "status" => "Hello",
        "in_reply_to_status_id" => activity.id
      })

    Tesla.Mock.mock(fn %{url: ^url, body: body} ->
      assert body =~ "mfc_id=1"
      assert body =~ "status_count=2"
      assert body =~ "last_post_id=#{reply.id}"
      assert body =~ "in_reply_to_id=#{activity.id}"
      send(self(), :called_reply_api)

      %Tesla.Env{status: 200}
    end)

    assert Notify.perform(:status_create, reply)
    assert_received(:called_reply_api)
  end

  describe "notify_user_change/1" do
    test "on follow to one side" do
      user1 = insert(:user, %{mfc_id: "1"})
      user2 = insert(:user, %{mfc_id: "2"})
      url = Config.get([:mfc, :user_info_sync_endpoint])

      {:ok, user1, user2, _} = CommonAPI.follow(user1, user2)

      Tesla.Mock.mock(fn %{url: ^url, body: body} ->
        assert body =~ "mfc_id=1"
        assert body =~ "friends=1"
        assert body =~ "followers=0"
        assert body =~ "active=true"
        send(self(), :called_api)
        %Tesla.Env{status: 200}
      end)

      assert Notify.perform(:user_change, user1)
      assert_received(:called_api)

      Tesla.Mock.mock(fn %{url: ^url, body: body} ->
        assert body =~ "mfc_id=2"
        assert body =~ "friends=0"
        assert body =~ "followers=1"
        assert body =~ "active=true"
        send(self(), :called_api)
        %Tesla.Env{status: 200}
      end)

      assert Notify.perform(:user_change, user2)
      assert_received(:called_api)
    end

    test "mutual folowing" do
      user1 = insert(:user, %{mfc_id: "1"})
      user2 = insert(:user, %{mfc_id: "2"})
      url = Config.get([:mfc, :user_info_sync_endpoint])

      {:ok, user1, user2, _} = CommonAPI.follow(user1, user2)
      {:ok, user2, user1, _} = CommonAPI.follow(user2, user1)

      Tesla.Mock.mock(fn %{url: ^url, body: body} ->
        assert body =~ "mfc_id=1"
        assert body =~ "friends=1"
        assert body =~ "followers=1"
        assert body =~ "active=true"
        send(self(), :called_api)
        %Tesla.Env{status: 200}
      end)

      assert Notify.perform(:user_change, user1)
      assert_received(:called_api)

      Tesla.Mock.mock(fn %{url: ^url, body: body} ->
        assert body =~ "mfc_id=2"
        assert body =~ "friends=1"
        assert body =~ "followers=1"
        assert body =~ "active=true"
        send(self(), :called_api)
        %Tesla.Env{status: 200}
      end)

      assert Notify.perform(:user_change, user2)
      assert_received(:called_api)
    end

    test "mutual folowing and unfollow" do
      user1 = insert(:user, %{mfc_id: "1"})
      user2 = insert(:user, %{mfc_id: "2"})

      url = Config.get([:mfc, :user_info_sync_endpoint])

      {:ok, user1, user2, _} = CommonAPI.follow(user1, user2)
      {:ok, user2, user1, _} = CommonAPI.follow(user2, user1)

      {:ok, user1} = CommonAPI.unfollow(user1, user2)

      Tesla.Mock.mock(fn %{url: ^url, body: body} ->
        assert body =~ "mfc_id=1"
        assert body =~ "friends=0"
        assert body =~ "followers=1"
        assert body =~ "active=true"
        send(self(), :called_api)
        %Tesla.Env{status: 200}
      end)

      assert Notify.perform(:user_change, user1)
      assert_received(:called_api)
    end

    test "accept follow request" do
      user1 = insert(:user, %{mfc_id: "1", info: %{locked: true}})
      user2 = insert(:user, %{mfc_id: "2"})
      url = Config.get([:mfc, :user_info_sync_endpoint])

      {:ok, user2, user1, _} = CommonAPI.follow(user2, user1)

      Tesla.Mock.mock(fn %{url: ^url, body: body} ->
        assert body =~ "mfc_id=1"
        assert body =~ "friends=0"
        assert body =~ "followers=0"
        assert body =~ "active=true"
        send(self(), :called_api)
        %Tesla.Env{status: 200}
      end)

      assert Notify.perform(:user_change, user1)
      assert_received(:called_api)

      Tesla.Mock.mock(fn %{url: ^url, body: body} ->
        assert body =~ "mfc_id=2"
        assert body =~ "friends=0"
        assert body =~ "followers=0"
        assert body =~ "active=true"
        send(self(), :called_api)
        %Tesla.Env{status: 200}
      end)

      assert Notify.perform(:user_change, user2)
      assert_received(:called_api)

      {:ok, user2} = CommonAPI.accept_follow_request(user2, user1)

      Tesla.Mock.mock(fn %{url: ^url, body: body} ->
        assert body =~ "mfc_id=2"
        assert body =~ "friends=1"
        assert body =~ "followers=0"
        assert body =~ "active=true"
        send(self(), :called_api)
        %Tesla.Env{status: 200}
      end)

      assert Notify.perform(:user_change, user2)
      assert_received(:called_api)

      user1 = Pleroma.User.get_by_id(user1.id)

      Tesla.Mock.mock(fn %{url: ^url, body: body} ->
        assert body =~ "mfc_id=1"
        assert body =~ "friends=0"
        assert body =~ "followers=1"
        assert body =~ "active=true"
        send(self(), :called_api)
        %Tesla.Env{status: 200}
      end)

      assert Notify.perform(:user_change, user1)
      assert_received(:called_api)
    end

    test "on deactivation" do
      user = insert(:user, %{mfc_id: "1"})
      url = Config.get([:mfc, :user_info_sync_endpoint])

      {:ok, user} = Pleroma.User.deactivate(user)

      Tesla.Mock.mock(fn %{url: ^url, body: body} ->
        assert body =~ "mfc_id=1"
        assert body =~ "friends=0"
        assert body =~ "followers=0"
        assert body =~ "active=false"
        send(self(), :called_api)
        %Tesla.Env{status: 200}
      end)

      assert Notify.perform(:user_change, user)
      assert_received(:called_api)
    end
  end
end
