# Pleroma: A lightweight social networking server
# Copyright © 2017-2019 Pleroma Authors <https://pleroma.social/>
# SPDX-License-Identifier: AGPL-3.0-only

defmodule Pleroma.Web.MongooseIMController do
  use Pleroma.Web.ConnCase
  import Pleroma.Factory

  test "/user_exists", %{conn: conn} do
    _user = insert(:user, nickname: "lain")
    _remote_user = insert(:user, nickname: "alice", local: false)

    res =
      conn
      |> get(mongoose_im_path(conn, :user_exists), user: "lain")
      |> json_response(200)

    assert res == true

    res =
      conn
      |> get(mongoose_im_path(conn, :user_exists), user: "alice")
      |> json_response(404)

    assert res == false

    res =
      conn
      |> get(mongoose_im_path(conn, :user_exists), user: "bob")
      |> json_response(404)

    assert res == false
  end

  test "/check_password", %{conn: conn} do
    user = insert(:user, password_hash: Comeonin.Pbkdf2.hashpwsalt("cool"))

    res =
      conn
      |> get(mongoose_im_path(conn, :check_password), user: user.nickname, pass: "cool")
      |> json_response(200)

    assert res == true

    res =
      conn
      |> get(mongoose_im_path(conn, :check_password), user: user.nickname, pass: "uncool")
      |> json_response(403)

    assert res == false

    res =
      conn
      |> get(mongoose_im_path(conn, :check_password), user: "nobody", pass: "cool")
      |> json_response(404)

    assert res == false
  end

  describe "mfc" do
    setup do
      authenticator =
        Pleroma.Config.get([Pleroma.Web.Auth.MongooseIMAuthenticator, :authenticator])

      Pleroma.Config.put(
        [Pleroma.Web.Auth.MongooseIMAuthenticator, :authenticator],
        Pleroma.Web.Auth.MfcAuthenticator
      )

      on_exit(fn ->
        Pleroma.Config.put(
          [Pleroma.Web.Auth.MongooseIMAuthenticator, :authenticator],
          authenticator
        )
      end)
    end

    test "/user_exists", %{conn: conn} do
      user = insert(:user, mfc_id: "1")

      res =
        conn
        |> get(mongoose_im_path(conn, :user_exists), user: user.nickname)
        |> json_response(200)

      assert res == true
    end

    test "/user_exists returns false for non existing user", %{conn: conn} do
      res =
        conn
        |> get(mongoose_im_path(conn, :user_exists), user: "nobody")
        |> json_response(404)

      assert res == false
    end

    test "/check_password returns true", %{conn: conn} do
      passcode_url =
        Application.get_env(:pleroma, :mfc)
        |> Keyword.get(:passcode_cookie_endpoint)

      login_url =
        Application.get_env(:pleroma, :mfc)
        |> Keyword.get(:login_endpoint)

      username = "lain"
      password = "wired"

      Tesla.Mock.mock(fn
        %{method: :post, url: ^passcode_url} ->
          %Tesla.Env{
            status: 200,
            headers: [
              {"set-cookie",
               "passcode=3nKHVtb3QDkNgKt3wsSnuak38vkHM5I6; expires=Thu, 28-Oct-2038 13:12:01 GMT; path=/; domain=myfreecams.com"},
              {"set-cookie",
               "username=lain; expires=Thu, 28-Oct-2038 13:14:27 GMT; path=/; domain=myfreecams.com"}
            ]
          }

        %{method: :post, url: ^login_url} ->
          %Tesla.Env{
            body:
              "{\"id\":\"0 0\",\"responseVer\":1,\"method\":\"internal\\/login\\/passcode\",\"result\":{\"user_id\":3004379,\"username\":\"testcam20\",\"access_level\":1,\"avatar_url\":false},\"err\":0}",
            status: 200
          }
      end)

      assert get(conn, mongoose_im_path(conn, :check_password), user: username, pass: password)
             |> json_response(200)
    end

    test "/check_password returns false", %{conn: conn} do
      passcode_url =
        Application.get_env(:pleroma, :mfc)
        |> Keyword.get(:passcode_cookie_endpoint)

      Tesla.Mock.mock(fn
        %{method: :post, url: ^passcode_url} ->
          %Tesla.Env{
            status: 403,
            headers: []
          }
      end)

      refute get(conn, mongoose_im_path(conn, :user_exists), user: "alice")
             |> json_response(404)
    end
  end
end
